from .display import greyscale
from .geometric import crop, rotate, flipping, reverse, scaling#, downsampling
from .histogram import adjustContrast, equalization, negative#, correctionGamma, histogram
from .filter import GaussianNoise, SaltnPepperNoise, filtering, sharpening, crosscorrelation
from .fourier import ftfiltering, UnsharpMasking, HighFreqEmph#, LaplacianFilter, hardthresholding



import math
import numpy
import scipy

from typing import Tuple, Any


def mse(x:numpy.ndarray, y: numpy.ndarray)->float:
    """
        Function that computes mse between two images.
        Input:
            x,y - images
        Output:
            computed mse
    """
    return 1/(x.shape[0]*x.shape[1]) * numpy.sum((x-y)**2)


def psnr(x:numpy.ndarray, y: numpy.ndarray)->float:
    """
        Function that computes psnr between two images.
        Input:
            x,y - images
        Output:
            computed psnr
    """
    return 10*numpy.log10(numpy.max(x)**2/mse(x,y))


def GaussianNoise(img:numpy.ndarray, mean:float, sigma:float)->numpy.ndarray:
    """
        Function that add gaussian noise to an image.
        Input:
            image - image 
            mean, sigma - parameters for gaussian distribution
        Output:
            image (with noise)
    """
    x = numpy.copy(img).astype(numpy.float32)
    noise = numpy.random.normal(mean, sigma,(x.shape[0],x.shape[1]))
    match len(x.shape):
        case 2:
            x += noise
        case 3:
            x[:,:,0] += noise
            x[:,:,1] += noise
            x[:,:,2] += noise
        case _:
            raise Exception("\n\033[{}m[-]Warning: Unknown size of image.".format("0;33"))
    x = numpy.clip(x, 0, 255).astype(numpy.uint8)
    return x


def SaltnPepperNoise(img:numpy.ndarray, prob:float)->numpy.ndarray:
    """
        Function that add gaussian noise to an image.
        Input:
            image - image
            prob - probability of salting
        Output:
            image (with noise)
    """
    x = numpy.copy(img)
    noise = numpy.random.random((x.shape[0],x.shape[1]))
    x[numpy.where(noise < prob)] = 0
    x[numpy.where(noise > 1-prob)] = 255
    return x



def filterLowPass()->numpy.ndarray:
    """
        Function that returns a lowpass filter.
        Input:
            None
        Output:
            filter
    other. 1/9*numpy.array([[1,1,1],[1,1,1],[1,1,1]])
    """
    return 1/8*numpy.array([[0,1,0],[1,4,1],[0,1,0]]) 


def filterHighPass()->numpy.ndarray:
    """
        Function that returns a highpass filter.
        Input:
            None
        Output:
            filter
    """
    return 1/4*numpy.array([[0,-1,0],[-1,8,-1],[0,-1,0]])
    #return numpy.array([[0,-1,0],[-1,4,-1],[0,-1,0]])
    #return 1/16*numpy.array([[-1,-2,-1],[-2,12,-2],[-1,-2,-1]])
    #return 1/9*numpy.array([[-1,-1,-1],[-1,8,-1],[-1,-1,-1]])
   

def filterBandreject()->numpy.ndarray:
    """
        Function that returns a band reject filter.
        Input:
            None
        Output:
            filter
    """
    return None

def filterBandpass()->numpy.ndarray:
    """
        Function that returns a bandpass filter.
        Input:
            None
        Output:
            filter
    """
    return None


def filterAverage(n:int)->numpy.ndarray:
    """
        Function that returns an average filter.
        Input:
            n - size of filter
        Output:
            filter
    """ 
    return 1/n**2*numpy.ones((n,n))
  


def filterGaussian()->numpy.ndarray: # sigma:float
    """
        Function that returns a gaussian filter.
        Input:
            sigma - parameter for Gaussian distribution
        Output:
            filter
    
        other.
            filter 1. - 1/1003*numpy.array([[0,0,1,2,1,0,0],[0,3,13,22,13,3,0],[1,13,59,97,59,13,1],[2,22,97,159,97,22,2],[1,13,59,97,59,13,1],[0,3,13,22,13,3,0],[0,0,1,2,1,0,0]])
            filter 2. - 1/273*numpy.array([[1,4,7,4,1],[4,16,26,16,4],[7,26,41,26,7],[4,16,26,16,4],[1,4,7,4,1]])
    """
    return 1/16*numpy.array([[1,2,1],[2,4,2],[1,2,1]])



def filterLaplacian()->numpy.ndarray:
    """
        Function that returns a Laplacian filter.
        Input:
            None
        Output:
            filter
        other. 
            filter . - numpy.array([[1,1,1],[1,-8,1],[1,1,1]])
    """
    return numpy.array([[0,1,0],[1,-4,1],[0,1,0]]) 


def filterSobel()->numpy.ndarray:
    """
        Function that returns a Sobel filter.
        Input:
            None
        Output:
            filter
        other. 
            filter . -numpy.array([[0,-1],[1,0]])
    """
    return numpy.array([[-1,0],[0,1]]) 
  
   

def filterRoberts()->numpy.ndarray:
    """
        Function that returns a Roberts filter.
        Input:
            None
        Output:
            filter
        other. 
            filter . - numpy.array([[-1,0,1],[-2,0,2],[-1,0,1]])
    """
    return numpy.array([[-1,-2,-1],[0,0,0],[1,2,1]])



def padding(img:numpy.ndarray, types:str, n:int)->numpy.ndarray:
    """
        Function to padd an image.
        Input:
            image - image
            types - types of padding {‘reflect’, ‘constant’, ‘wrap’...‘nearest’, ‘mirror’,}
            n - size of padding
        Output:
            image (padded)
    """
    # A TERMINER !!! - add different types of padding

    size = numpy.copy(img.shape)
    size[0] = 3*img.shape[0]
    size[1] = 3*img.shape[1]
    x = numpy.zeros(size)

    match types:
        case "constant":
            """
            size = numpy.copy(img.shape)
            size[0] = img.shape[0]+2*n
            size[1] = img.shape[1]+2*n
            x = numpy.zeros(size)
            x[n:img.shape[0]+n, n:img.shape[1]+n] = img
            """
            x[img.shape[0]:2*img.shape[0], img.shape[1]:2*img.shape[1]] = img
        case "wrap":
            for i in range(3):
                for j in range(3):
                    x[i*img.shape[0]:(i+1)*img.shape[0], j*img.shape[1]:(j+1)*img.shape[1]] = img
        case "reflect":
            x[0:img.shape[0], 0:img.shape[1]] = (img[:, ::-1])[::-1, :]
            x[0:img.shape[0], img.shape[1]:2*img.shape[1]] = (img)[::-1, :]
            x[0:img.shape[0], 2*img.shape[1]:] = (img[:, ::-1])[::-1, :]
            x[img.shape[0]:2*img.shape[0], 0:img.shape[1]] = img[:, ::-1]
            x[img.shape[0]:2*img.shape[0], img.shape[1]:2*img.shape[1]] = img
            x[img.shape[0]:2*img.shape[0], 2*img.shape[1]:] = img[:, ::-1]
            x[2*img.shape[0]:, 0:img.shape[1]] = (img[:, ::-1])[::-1, :]
            x[2*img.shape[0]:, img.shape[1]:2*img.shape[1]] = (img[::-1, ::])
            x[2*img.shape[0]:, 2*img.shape[1]:] = (img[:, ::-1])[::-1, :]
        case "nearest":
            pass
        case "mirror":
            pass
        case _:
            raise Exception("\n\033[{}m[-]Warning: Unknown type of padding.".format("0;33"))
    x = x[img.shape[0]-n:2*img.shape[0]+n, img.shape[1]-n:2*img.shape[1]+n]
    x = numpy.clip(x, 0, 255).astype(numpy.uint8)
    return x

def convolution(img:numpy.ndarray, filters:numpy.ndarray, mode:str)->numpy.ndarray:
    """
        Function that convolves two images.
        Input:
            img - image
            filters - filter to apply
            mode - type of padding
        Output:
            image (convolution)
    """
    padds = math.ceil((filters.shape[0]-1)/2)
    x = padding(img, mode,padds)
   
    n = img.shape[0]+2*padds-filters.shape[0]+1
    m = img.shape[1]+2*padds-filters.shape[0]+1
    match len(x.shape):
        case 2:
            result = numpy.zeros((n, m))
        case 3:
            result = numpy.zeros((n, m, x.shape[2]))
            if len(filters.shape)==2:
                new = numpy.zeros((filters.shape[0],filters.shape[1], x.shape[2]))
                for i in range(x.shape[2]):
                    new[:,:,i] = filters 
                filters = new               
        case _:
            raise Exception("\n\033[{}m[-]Warning: Unknown size of image.".format("0;33"))

    for i in range(n):
        for j in range(m):
            z = x[i:i+filters.shape[0], j:j+filters.shape[1]] 
            z = z[:, ::-1]
            z = z[::-1, :]
            convolv = z * filters 

            result[i,j] = numpy.sum(convolv)
        
    result = result[:img.shape[0], :img.shape[1]] # crop image, because of padding
    return result



def filtering(img:numpy.ndarray, types:str, mode:str, *args: Tuple[Any])->numpy.ndarray:
    """
        Function that apply filter on an image.
        Input:
            img - image on which th filter is applied
            types - type of filter 
            mode - type of padding {‘reflect’, ‘constant’, ‘nearest’, ‘mirror’, ‘wrap’}
            args - extra arguments (optional)
        Output:
            image (filtered)
    """
    match types:
        case "lowpass":
            filters = filterLowPass()
        case "highpass":
            filters = filterHighPass()
        case "bandreject":
            filters = filterBandreject()
        case "bandpass":
            filters = filterBandpass()
        case "avg":
            filters = filterAverage(int(args[0]))
        case "gauss":
            filters = filterGaussian() #int(args[0]))
        case _:
            raise Exception("\n\033[{}m[-]Warning: Unknown type of filter.".format("0;33"))

    #x = convolution(img, filters, mode)

    x = numpy.zeros(img.shape)
    match len(img.shape):
        case 2:
            x = scipy.ndimage.convolve(img, filters, mode=mode)
        case 3:
            for i in range(img.shape[2]):
                x[:,:,i] = scipy.ndimage.convolve(img[:,:,i], filters, mode=mode)
        case _:
            raise Exception("\n\033[{}m[-]Warning: Unknown size of image.".format("0;33"))
  
    x = numpy.clip(x, 0, 255).astype(numpy.uint8)
    return x



def sharpening(img:numpy.ndarray, types:str, mode:str)->numpy.ndarray:
    """
        Function that performs sharpening on an image.
        Input:
            img - image on which th filter is applied
            types - type of filter 
            mode - type of padding {‘reflect’, ‘constant’, ‘wrap’...‘nearest’, ‘mirror’,}
        Output:
            image (filtered)
    """
    match types:
        case "laplacian":
            filters = filterLaplacian()
        case "sobel":
            filters = filterSobel()
        case "roberts":
            filters = filterRoberts()
        case _:
            raise Exception("\n\033[{}m[-]Warning: Unknown type of filter.".format("0;33"))
    x = convolution(img, filters, mode)
    x = numpy.clip(x, 0, 255).astype(numpy.uint8)
    return x


def crosscorrelation(img:numpy.ndarray, filters:str, sub:numpy.ndarray)->numpy.ndarray:
    """
        Function that find a subimage in an image.
        Input:
            img - image
            sub - subimage to find
        Output:
            subimage (found)
    """
    # A TERMINER !!! - detect elements of same shape but different colors
    
    original = numpy.copy(img)

    img = 0.2989 * img[:,:,0] + 0.5870 * img[:,:,1] + 0.1140 * img[:,:,2]
    sub = 0.2989 * sub[:,:,0] + 0.5870 * sub[:,:,1] + 0.1140 * sub[:,:,2]
    img = sharpening(img, filters, "reflect") 
    sub = sharpening(sub, filters, "reflect") 

    # padding

    size = numpy.copy(img.shape)
    size[0] = img.shape[0]-sub.shape[0]+1
    size[1] = img.shape[1]-sub.shape[0]+1
    result = numpy.zeros(size)

    index = (0,0)
    for i in range(result.shape[0]):
        for j in range(result.shape[1]):
            corr = img[i:i+sub.shape[0], j:j+sub.shape[1]] * sub
            result[i,j] = numpy.sum(corr)
            
            if numpy.max(result[index]) < numpy.max(result[i,j]):
                index = (i,j)  
    
    N = 5
    COLOR = numpy.array([255,0,127])
    
    find = numpy.copy(original[index[0]:index[0]+sub.shape[0], index[1]:index[1]+sub.shape[1]])    
           
    original[index[0]-N:index[0]+sub.shape[0]+N, index[1]-N:index[1]+sub.shape[1]+N] = COLOR 
    original[index[0]:index[0]+sub.shape[0], index[1]:index[1]+sub.shape[1]] = find        
    return original 
    #return original[index[0]:index[0]+sub.shape[0], index[1]:index[1]+sub.shape[1]] 













import numpy
import matplotlib.pyplot as plt


def histogram(img:numpy.ndarray, density:bool,)->None:
    """
        Function that plots histogram of an image.
        Input:
            img - image
        Output:
            None
    """
    # hist, bins = numpy.histogram(img, bins=numpy.arange(256))#, bins='auto', density=True)
    match len(img.shape):
        case 2:
            unique, counts = numpy.unique(img, return_counts=True)
            plt.hist(counts, bins=unique, density=density)
        case 3:
            fig, ax = plt.subplots(3, 1)
            channels = ["Red", "Green", "Blue"]
            for i in range(img.shape[2]):
                unique, counts = numpy.unique(img[:,:,i], return_counts=True)
                ax[i].hist(counts, bins=unique, density=density) 
                ax[i].set_title("Channel {}".format(channels[i]))
        case _:
            raise Exception("\n\033[{}m[-]Warning: Unknown size of image.".format("0;33"))
    plt.show()
 



def adjustContrast(img:numpy.ndarray)->numpy.ndarray: # A REVOIR !!!
    """
        Function that performs contrast stretching.
        Input:
            img - image
        Output:
            image (with contrat stretching)
    """
    f = lambda x, h, H : 255/(H-h)*(x-h)
    x = numpy.zeros(img.shape)
    match len(img.shape):
        case 2:
            unique, counts = numpy.unique(img, return_counts=True)
            x = f(img, numpy.min(unique), numpy.max(unique))
        case 3:
            for i in range(img.shape[2]):
                unique, counts = numpy.unique(img[:,:,i], return_counts=True)
                x[:,:,i] = f(img[:,:,i], numpy.min(unique), numpy.max(unique))  
            """
            unique, counts = numpy.unique(img, return_counts=True)
            x = f(img, numpy.min(unique), numpy.max(unique)) 
            """        
        case _:
            raise Exception("\n\033[{}m[-]Warning: Unknown size of image.".format("0;33"))
    x = numpy.clip(x, 0, 255).astype(numpy.uint8)
    return x


def equalization(img:numpy.ndarray)->numpy.ndarray: # A REVOIR !!!
    """
        Function that performs histogram equalization.
        Input:
            img - image
        Output:
            image (with equalization)
    """
    x = numpy.copy(img)
    match len(img.shape):
        case 2:
            unique, counts = numpy.unique(img, return_counts=True)
            const = (numpy.min(unique)-1)/(img.shape[0]*img.shape[1])
            for i in range(unique.shape[0]):
                   x[numpy.where(x==unique[i])] = const*numpy.sum(counts[:i])
        case 3:
            unique, counts = numpy.unique(img, return_counts=True)
            const = (numpy.min(unique)-1)/(img.shape[0]*img.shape[1])
            for i in range(unique.shape[0]):
                x[numpy.where(x==unique[i])] = const*numpy.sum(counts[:i])     
        case _:
            raise Exception("\n\033[{}m[-]Warning: Unknown size of image.".format("0;33"))
    x = numpy.clip(x, 0, 255).astype(numpy.uint8)
    return x

def correctionGamma(img:numpy.ndarray, c:float, gamma: float)->numpy.ndarray:
    """
        Function that performs gamma correction on an image.
        Input:
            img - image (grey image)
            c, gamma - parameters of formula
        Output:
            image (gamma correction)
        other.
             s = T(r) = c*r**gamma , c = 1
    """
    f = lambda x: c*x**gamma
    x = f(img)
    x = numpy.clip(x, 0, 255).astype(numpy.uint8)
    return x

def negative(img:numpy.ndarray)->numpy.ndarray:
    """
        Function that computes digital negative of an image.
        Input:
            img - image
        Output:
            image (negative)
        other.
            s = T(r) = (L – 1) – r
    """
    f = lambda x, m : (m-1)-x
    x = numpy.zeros(img.shape)
    match len(img.shape):
        case 2:
            unique, counts = numpy.unique(img, return_counts=True)
            x = f(img, numpy.min(unique))
        case 3:
            for i in range(img.shape[2]):
                unique, counts = numpy.unique(img[:,:,i], return_counts=True)
                x[:,:,i] = f(img[:,:,i], numpy.min(unique))          
        case _:
            raise Exception("\n\033[{}m[-]Warning: Unknown size of image.".format("0;33"))
    x = numpy.clip(x, 0, 255).astype(numpy.uint8)
    return x





import numpy
import scipy
import matplotlib.pyplot as plt

from typing import Tuple, Any


def Fourier(img:numpy.ndarray)->None:
    """
        Function that computes Fourier transform of an image.
        Input:
            img - image
        Output:
            None
    """
    x = scipy.fft.fftn(img)
    x = scipy.fft.fftshift(x)

    # plot phase and magnitude
    magnitude = numpy.abs(x) #numpy.sqrt(xreal**2 + ximag**2) 
    phase = numpy.arctan2(numpy.imag(x),numpy.real(x))  # numpy.angle(x)

    # reconstruct image
    y = magnitude * numpy.exp(1j*phase)
    y = scipy.fft.ifftshift(y)
    y = scipy.fft.ifftn(y) 
    y = numpy.real(y)
    y = numpy.clip(y, 0, 255).astype(numpy.uint8)
    plt.imshow(y)
    plt.show()



def hardthresholding(img:numpy.ndarray, threshold:float)->numpy.ndarray:
    """
        Function that performs hard thresholding.
        Input:
            img - image
            threshold - 
        Output:     
            image (threshold)
    """
    x = scipy.fft.fftn(img)
    #x = scipy.fft.fftshift(x)

    magnitude = numpy.abs(x)
    phase = numpy.arctan2(numpy.imag(x),numpy.real(x)) # numpy.angle(x)

    threshold *= numpy.mean(magnitude)
    x[numpy.where(magnitude < threshold)] = 0 #numpy.where(magnitude > threshold)

    magnitude = numpy.abs(x)
    phase = numpy.arctan2(numpy.imag(x),numpy.real(x))

    y = magnitude * numpy.exp(1j*phase)
    #y = scipy.fft.ifftshift(y)
    y = scipy.fft.ifftn(x) 
    y = numpy.real(y)
    y = numpy.clip(y, 0, 255).astype(numpy.uint8)
    return y



def IdealHighPassFilter(img:numpy.ndarray, D0:float)->numpy.ndarray:
    """
        Function that returns an Ideal Lowpass filter.
        Input:
            img - image
            D0 - cutoff frequency
        Output:
            filter
    """
    P = img.shape[0]
    Q = img.shape[1]
    D = lambda u,v: numpy.sqrt((u-P/2)**2 + (v-Q/2)**2)  

    H = numpy.zeros((P,Q))
    for u in range(P):
        for v in range(Q):
            if D(u,v) <= D0:
                H[u,v] = 1
            else:
                H[u,v] = 0
    return H

def ButterworthHighPassFilter(img:numpy.ndarray, D0:float, n:int)->numpy.ndarray:
    """
        Function that returns a Butterworth Lowpass filter.
        Input:
        Input:
            img - image
            D0 - cutoff frequency
            n - order
        Output:
            filter
    """
    P = img.shape[0]
    Q = img.shape[1]
    D = lambda u,v: numpy.sqrt((u-P/2)**2 + (v-Q/2)**2)  
    Butterworth = lambda u,v: 1/(1 + (D(u,v)/D0)**(2*n))

    H = numpy.zeros((P,Q))
    for u in range(P):
        for v in range(Q):
           H[u,v] = Butterworth(u,v)
    return H

def GaussianHighPassFilter(img:numpy.ndarray, sigma:float)->numpy.ndarray:
    """
        Function that returns an Gaussian Lowpass filter.
        Input:
            img - image
            sigma - cutoff frequency
        Output:
            filter
    """
    P = img.shape[0]
    Q = img.shape[1]
    D = lambda u,v: numpy.sqrt((u-P/2)**2 + (v-Q/2)**2)  
    gaussian = lambda u,v: numpy.exp(-D(u,v)**2/(2*sigma**2))

    H = numpy.zeros((P,Q))
    for u in range(P):
        for v in range(Q):
           H[u,v] = gaussian(u,v)
    return H

def IdealLowPassFilter(img:numpy.ndarray, D0:float)->numpy.ndarray:
    """
        Function that returns an Ideal Highpass filter.
        Input:
            img - image
            D0 - cutoff frequency
        Output:
            filter
    """
    P = img.shape[0]
    Q = img.shape[1]
    D = lambda u,v: numpy.sqrt((u-P/2)**2 + (v-Q/2)**2) 

    H = numpy.zeros((P,Q))
    for u in range(P):
        for v in range(Q):
            if D(u,v) <= D0:
                H[u,v] = 0
            else:
                H[u,v] = 1
    return H

def ButterworthLowPassFilter(img:numpy.ndarray, D0:float, n:int)->numpy.ndarray:
    """
        Function that returns a Butterworth Highpass filter.
        Input:
        Input:
            img - image
            D0 - cutoff frequency
            n - order
        Output:
            filter
    """
    P = img.shape[0]
    Q = img.shape[1]
    D = lambda u,v: numpy.sqrt((u-P/2)**2 + (v-Q/2)**2)  
    Butterworth = lambda u,v: 1/(1 + (D0/D(u,v))**(2*n))

    H = numpy.zeros((P,Q))
    for u in range(P):
        for v in range(Q):
           H[u,v] = Butterworth(u,v)
    return H


def GaussianLowPassFilter(img:numpy.ndarray, sigma:float)->numpy.ndarray:
    """
        Function that returns an Gaussian Highpass filter.
        Input:
            img - image
            sigma - cutoff frequency
        Output:
            filter
    """
    P = img.shape[0]
    Q = img.shape[1]
    D = lambda u,v: numpy.sqrt((u-P/2)**2 + (v-Q/2)**2) 
    gaussian = lambda u,v: numpy.exp(-D(u,v)**2/(2*sigma**2))

    H = numpy.zeros((P,Q))
    for u in range(P):
        for v in range(Q):
           H[u,v] = 1-gaussian(u,v)
    return H


def HomomorphicFilter(img:numpy.ndarray, D0:int, Hgamma:float, Lgamma:float, c:float)->numpy.ndarray:
    """
        Function that returns a homorphic filter.
        Input:
            img - image
            D0 - cutoff frequency
            c - constant
            gammah, gammal - dynamic range (gammaH, gammaL)
        Output:
            filter
        Other.
            H(u,v) = (gammaH - gammaL) (1 - exp(-(c*D(u,v)**2) / D0**2)) + gammaL
    """
    P = img.shape[0]
    Q = img.shape[1]
    D = lambda u,v: numpy.sqrt((u-P/2)**2 + (v-Q/2)**2)
    fgauss = lambda u, v: numpy.exp(-c*D(u,v)**2 / D0**2)

    H = numpy.zeros((P,Q))
    for u in range(P):
        for v in range(Q):
           H[u,v] = (Hgamma-Lgamma) * (1-fgauss(u,v)) + Lgamma
    return H


def BandrejectFilter(img:numpy.ndarray, D0:float, W:float, filters:str, args:Tuple[Any])->numpy.ndarray:
    """
        Function that returns bandreject filter.
        Input:
            img - image
            D0 - cutoff frequency
            W - width of the band
            filters - type of filter
            *args - parameters (optional)
        Output:
            filter
    """
    P = img.shape[0]
    Q = img.shape[1]
    D = lambda u,v: numpy.sqrt((u-P/2)**2 + (v-Q/2)**2)  

    H = numpy.zeros((P,Q))

    match filters:
        case "ideal":
            for u in range(P):
                for v in range(Q):
                    if D0-W/2<=D(u, v) and D(u,v)<=D0+W/2:
                        H[u,v] = 0
                    else:
                        H[u,v] = 1
        case "butterworth":
            n = int(args[0])
            for u in range(P):
                for v in range(Q):
                    H[u,v] = 1 / (1 + (D(u,v)*W /(D(u,v)**2-D0**2))**(2*n) )
        case "gauss":
            for u in range(P):
                for v in range(Q):
                    H[u,v] = 1 - numpy.exp(- ((D(u,v)**2-D0**2) / (D(u,v)*W))**2)
        case _:
            raise Exception("\n\033[{}m[-]Warning: Unknown type of filter.".format("0;33"))
    return H


def BandpassFilter(img:numpy.ndarray, D0:float, W:float, filters:str, args:Tuple[Any])->numpy.ndarray:
    """
        Function that returns bandpass filter.
        Input:
            img - image
            D0 - cutoff frequency
            W - width of the band
            filters - type of filter
            *args - parameters (optional)
        Output:
            filter
    """
    return 1-BandrejectFilter(img, D0, W, filters, args) # numpy.abs()



def LaplacianFilter(img:numpy.ndarray, c:int)->numpy.ndarray:
    """
        Function that returns an Gaussian Laplacian filter.
        Input:
            img - image
            c - constant
        Output:
            filter
    """
    P = img.shape[0]
    Q = img.shape[1]
    D = lambda u,v: numpy.sqrt((u-P/2)**2 + (v-Q/2)**2)  

    H = numpy.zeros((P,Q))
    for u in range(P):
        for v in range(Q):
           H[u,v] = -D(u,v)**2
    
    match len(img.shape):
        case 2:
            pass
        case 3:
            h = numpy.zeros((H.shape[0], H.shape[1], img.shape[2]))
            for i in range(img.shape[2]):
                h[:,:, i] = H  
            H = h             
        case _:
            raise Exception("\n\033[{}m[-]Warning: Unknown size of image.".format("0;33"))
    F = scipy.fft.fftn(img)
    df = numpy.abs(H)*numpy.abs(F)* numpy.exp(1j*numpy.angle(F))
    df = scipy.fft.ifftn(df) 
    df = numpy.real(df)
    g = img + c*df
    #g = numpy.clip(g, 0, 255).astype(numpy.uint8)
    return g


def UnsharpMasking(img:numpy.ndarray, D0:int, k:int)->numpy.ndarray: 
    """
        Function that performs Unsharp Masking and Highboost filtering.
        Input:
            img - image
            k - parameter
                    k=1 - unsharp masking
                    k>1 - highboost filtering
        Output:
            image
    """
    F = scipy.fft.fftn(img)
    H = IdealLowPassFilter(img, D0) # IdealHighPassFilter(img, D0)
    match len(img.shape):
        case 2:
            pass
        case 3:
            h = numpy.zeros((H.shape[0], H.shape[1], img.shape[2]))
            for i in range(img.shape[2]):
                h[:,:, i] = H  
            H = h             
        case _:
            raise Exception("\n\033[{}m[-]Warning: Unknown size of image.".format("0;33"))
    G = numpy.abs(H)*numpy.abs(F)* numpy.exp(1j*numpy.angle(F))
    gmask = scipy.fft.ifftn(G) 
    gmask = numpy.real(gmask)
    #gmask = numpy.clip(gmask, 0, 255).astype(numpy.uint8)
    g = img + k * gmask
    g = numpy.clip(g, 0, 255).astype(numpy.uint8)
    return g


def HighFreqEmph(img:numpy.ndarray, D0:int, k1:int, k2:int)->numpy.ndarray: 
    """
        Function that performs High-frequency emphasis.
        Input:
            img - image
            k1, k2 - parameter
                    k1 = 1
                    k2=1 - unsharp masking
                    k2>1 - highboost filtering
        Output:
            image
    """
    F = scipy.fft.fftn(img)
    H = IdealLowPassFilter(img, D0) # IdealHighPassFilter(img, D0)
    match len(img.shape):
        case 2:
            pass
        case 3:
            h = numpy.zeros((H.shape[0], H.shape[1], img.shape[2]))
            for i in range(img.shape[2]):
                h[:,:, i] = H  
            H = h             
        case _:
            raise Exception("\n\033[{}m[-]Warning: Unknown size of image.".format("0;33"))
    H = k1 +k2*H
    G = numpy.abs(H)*numpy.abs(F)* numpy.exp(1j*numpy.angle(F))
    y = scipy.fft.ifftn(G) 
    y = numpy.real(y)
    y = numpy.clip(y, 0, 255).astype(numpy.uint8)
    return y


def NotchFilter()->numpy.ndarray:
    """
        Function that ...
        Input:
        Output:
    """
    return None


def ftfiltering(img:numpy.ndarray, filters:str, *args: Tuple[Any])->numpy.ndarray: # mode:str
    """
        Function that performs filtering.
        Input:
            img - image
            filters - type of filter {lowpass, blowpass, glowpass, highpass, bhighpass, ghighoass, bandreject, bandpass,homomorphic}
            *args - (opt.)
        Output:
            image (filtered)
    """
    # TERMINER !!! - case of padding (image)
    match filters:
        case "lowpass":
            H = IdealLowPassFilter(img, float(args[0]))
        case "blowpass":
            H = ButterworthLowPassFilter(img, float(args[0]), float(args[1]))
        case "glowpass":
            H = GaussianLowPassFilter(img, float(args[0]))
        case "highpass":
            H = IdealHighPassFilter(img, float(args[0]))
        case "bhighpass":
            H = ButterworthHighPassFilter(img, float(args[0]), float(args[1]))
        case "ghighpass":
            H = GaussianHighPassFilter(img, float(args[0]))
        
        case "bandreject":
            H = BandrejectFilter(img, float(args[1]), float(args[2]), args[0], args[3:])
        case "bandpass":
            H = BandpassFilter(img, float(args[1]), float(args[2]), args[0], args[3:])

        case "homomorphic":
            H = HomomorphicFilter(img, float(args[0]), float(args[1]), float(args[2]), float(args[3]))
           
        case _:
            raise Exception("\n\033[{}m[-]Warning: Unknown type of filter.".format("0;33"))

    match len(img.shape):
        case 2:
            pass
        case 3:
            h = numpy.zeros((H.shape[0], H.shape[1], img.shape[2]))
            for i in range(img.shape[2]):
                h[:,:, i] = H  
            H = h             
        case _:
            raise Exception("\n\033[{}m[-]Warning: Unknown size of image.".format("0;33"))

    F = scipy.fft.fftn(img)
    #F = scipy.fft.fftshift(F)

    Fphase = numpy.arctan2(numpy.imag(F),numpy.real(F))

    # perform filtering
    G = numpy.abs(H)*numpy.abs(F)* numpy.exp(1j*numpy.angle(F))

    #y = scipy.fft.ifftshift(G)
    y = scipy.fft.ifftn(G) 
    y = numpy.real(y)
    y = numpy.clip(y, 0, 255).astype(numpy.uint8)
    return y



def ftcrosscorrelation(img:numpy.ndarray)->numpy.ndarray:
    """
        Function that ...
        Input:
        Output:
    """
    return img







BLACK = "1;30"
RED = "1;31"
GREEN = "1;32"
YELLOW = "1;33"
BLUE = "1;34"
MAGENTA = "1;35"
CYAN = "1;36"
WHITE = "1;37"


# File display.py
WARNINGFILE = "\n\033[{}m[-]Error: File does not exist.".format("0;31")

# File transform.py
WARNINGCROPPING = "\n0\33[{}m[-]Error: Impossible to crop image.".format("0;31")

# File filter.py
WARNINGIMAGESIZE ="\n\033[{}m[-]Warning: Unknown size of image.".format("0;33")
WARNINGPADDING = "\n\033[{}m[-]Warning: Unknown type of padding.".format("0;33")
WARNINGFILTER = "\n\033[{}m[-]Warning: Unknown type of filter.".format("0;33")

# FIle FT.py
WARNINGFTFILTER = "\n\033[{}m[-]Warning: Unknown type of filter.".format("0;33")









import os
import sys
import numpy
import matplotlib.pyplot as plt
import matplotlib.image as mpimg


def greyscale(x: numpy.ndarray)->numpy.ndarray:
    """
        Function that converts image into greyscale.
        Input:
            x - image to convert
        Output:
            image (greyscale)
    """
    R, G, B = x[:, :, 0], x[:, :, 1], x[:, :, 2]
    y = 0.2989 * R + 0.5870 * G + 0.1140 * B
    return y


def displayRGB(x: numpy.ndarray)->None:
    """
        Function that displays R,G,B channels of an image.
        Input:
            x - image to display
        Output:
            None
    """
    fig, ax = plt.subplots(1, 3)
    cmaps = [plt.cm.Reds_r, plt.cm.Greens_r, plt.cm.Blues_r]
    channels = ["Red", "Green", "Blue"]
    for i in range(x.shape[2]):
        ax[i].imshow(x[:,:, i], cmap=cmaps[i]) 
        ax[i].set_title("Channel {}".format(channels[i]))
    plt.show()
 

def display(x: numpy.ndarray, y:str)->None:
    """
        Function that displays an image.
        Input:
            x - image to display
        Output:
            None
    """
    plt.title(y)
    plt.xlabel("X pixels scaling")
    plt.ylabel("Y pixels scaling")
    match len(x.shape):
        case 2:
            plt.imshow(x,  cmap='gray')
        case 3:
            plt.imshow(x)
        case _:
            raise Exception("\n\033[{}m[-]Warning: Unknown size of image.".format("0;33")) 
    plt.show()





    
    

        



import sys
import math
import numpy

from typing import Tuple



def crop(img:numpy.ndarray, x:Tuple[float], width:int, height:int)->numpy.ndarray:
    """
        Function that crops an image.
        Input:
            img - image to crop
            x - point
            width - width of the cropped image
            height - height of the cropped image
        Output:
    """
    y = numpy.copy(img)    
    if not(x[0]>=0 and x[0]<img.shape[0] and x[1]>=0 and x[1]<img.shape[1]):
        raise Exception("\n0\33[{}m[-]Error: Impossible to crop image.".format("0;31"))
        sys.exit()
    if (x[0] + width > img.shape[0] or x[1] + height > img.shape[1]):
        raise Exception("\n0\33[{}m[-]Error: Impossible to crop image.".format("0;31"))
        sys.exit()
    return y[x[0]:x[0]+width, x[1]:x[1]+height]



def compression(img:numpy.ndarray, fact:float)->numpy.ndarray:
    """
        Function that ...
        Input:
        Output:
    """
    return img


def downsampling(img:numpy.ndarray, fact:float)->numpy.ndarray:
    """
        Function that performs downsampling on an image.
        Input:
            img - image to downsample
            fact - factor for downsampling
        Output:
            image (downsampled)
    """
    match len(img.shape):
        case 2:
            return img[::fact, ::fact]
        case 3:
            return img[::fact, ::fact, :]
        case _:
            raise Exception("\n\033[{}m[-]Warning: Unknown size of image.".format("0;33"))                                                                                                                        


def upsampling(img:numpy.ndarray, fact:float)->numpy.ndarray:
    """
        Function that ...
        Input:
        Output:
    """
    # nearest interpolation
    # bilinear interpolation
    return img


def scaling(img:numpy.ndarray, scale:float)->numpy.ndarray:
    """
        Function that scales an image.
        Input:
            img - image to scale
            scale - factor of scaling
        Output:
            image (scaled)
    """
    size = numpy.copy(img.shape)
    size[0] = int(img.shape[0]*scale)
    size[1] = int(img.shape[1]*scale)
    x = numpy.zeros(size)
    for i in range(size[0]):
        for j in range(size[1]):
            x[i,j] = img[int(i/scale), int(j/scale)]        
    x = numpy.clip(x, 0, 255).astype(numpy.uint8)
    return x



def rotate(img:numpy.ndarray, degree:float)->numpy.ndarray:
    """
        Function that rotates an image.
        Input:
            img - image to rotate
            degree - angle of rotation (degrees)
        Output:
            image (rotate)
    """
    rads = math.radians(degree)
    result = numpy.zeros(img.shape)

    midx,midy = (result.shape[0]//2, result.shape[1]//2)

    for i in range(result.shape[0]):
        for j in range(result.shape[1]):
            x = (i-midx)*math.cos(rads)+(j-midy)*math.sin(rads)
            y = -(i-midx)*math.sin(rads)+(j-midy)*math.cos(rads)

            x = round(x)+midx
            y = round(y)+midy 
            
            if (x>=0 and y>=0 and x<img.shape[0] and  y<img.shape[1]):
                result[i,j,:] = img[x,y,:]
    result = numpy.clip(result, 0, 255).astype(numpy.uint8)
    return result


def flipping(img:numpy.ndarray)->numpy.ndarray:
    """
        Function that flips an image.
        Input:
            img - image to flip
        Output:
            image (flip)
    """
    return img[:, ::-1, :]

def reverse(img:numpy.ndarray)->numpy.ndarray:
    """
        Function that reverse an image.
        Input:
            img - image to reverse
        Output:
            image (reverse)
    """
    return img[::-1, :, :]


def translation(img:numpy.ndarray)->numpy.ndarray:
    """
        Function that ...
        Input:
        Output:
    """
    return img

def projection(img:numpy.ndarray)->numpy.ndarray:
    """
        Function that ...
        Input:
        Output:
    """
    return img






from setuptools import setup, find_packages



setup(
	name = "imgmagic",
	version = "0.1.0",
	packages = find_packages(),
	install_requires = [
		'numpy==1.23.4',
		'scipy==1.11.4',
		'matplotlib==3.5.1',
		'pillow==9.0.1', 
		'protobuf<4'
	],
)
